game = {
	levelFile = "levels/laserlev.bin",
	levelNumber = 1,
	levelOffset = 0,
	quit = false,
	load = function()
		game.digits = image.load("images/time-digits.png")
	end,
	close = function()
		game.digits:close()
	end,
	drawControls = function()
		
		-- Draw the control panel to the right of the screen.
		screen.fillrect(0, 0, 400, 240, color.new(192, 192, 192))
		screen.fillrect(360, 0, 2, 240, color.new(255, 255, 255))
		
		-- Draw the clock space.
		screen.fillrect(366, 3, 30, 71, color.new(0, 0, 0))
		game.drawDigit(10, 368, 32)
		
		-- Draw its buttons.
		img = image.load("images/load.png")
		img:draw(366, 78)
		img:close()
		
		img = image.load("images/restart.png")
		img:draw(366, 159)
		img:close()
	end,
	start = function()
		
		-- Load the level.
		level.clearTemporaryTiles()
		level.load(game.levelFile, game.levelNumber)
		
		-- Draw the controls to the right of the display.
		game.drawControls()
		
		-- Draw the level.
		level.draw()
		
		-- Stop the laser.
		laser.active = false
		
		-- Start the timer.
		local delaySecond = os.time()
		while os.time() == delaySecond do end
		game.startTime = os.time()
		game.drawClock()
		game.pausedTimer = nil
		
		-- Update the screen.
		screen.update()
		
	end,
	drawDigit = function(digit, x, y)
		game.digits:draw(x, y - digit * 13, x, y, 25, 13)
	end,
	-- Show the timer.
	drawClock = function()
		
		-- How long have we been playing?
		local seconds = os.time() - game.startTime
		game.lastDrawnClockSeconds = seconds
		
		-- Fix the number of seconds.
		if seconds < 0 then seconds = 0 end
		
		-- Calculate the number of minutes.
		local minutes = math.floor(seconds / 60)
		
		-- Have we been playing for over 99 minutes?
		if minutes > 99 then
			minutes = 99
			seconds = 59
		else
			seconds = seconds % 60
		end
		
		-- Draw the digits.
		game.drawDigit(seconds % 10, 368, 6)
		game.drawDigit(math.floor(seconds / 10), 368, 21)
		game.drawDigit(minutes % 10, 368, 43)
		game.drawDigit(math.floor(minutes / 10), 368, 58)
		
	end,
	pauseTimer = function()
		if not game.pausedTimer then
			local delaySecond = os.time()
			while os.time() == delaySecond do end
			game.pausedTimer = os.time()
		end
	end,
	resumeTimer = function()
		if game.pausedTimer then
			local delaySecond = os.time()
			while os.time() == delaySecond do end
			local pausedTime = os.time() - game.pausedTimer + 1
			if pausedTime > 0 then game.startTime = game.startTime + pausedTime end
			game.lastUpdateTime = nil
			game.pausedTimer = nil
		end
	end,
	run = function()
		-- Run the main game loop.
		while true do
			-- Attempt to get input.
			local hasInput = control.read() == 1
			if hasInput then
				if control.isButton() == 1 then
					if button.click() == 1 and not laser.active then
						-- We've clicked the button; try to quit.
						sounds.play("clck")
						game.pauseTimer()
						if dialog.show("end-game-confirmation", { "ok", "cancel" }, "question") == "ok" then
							break
						else
							game.resumeTimer()
							-- Redraw the level to hide the dialog box.
							level.draw()
						end
					elseif button.hold() == 1 then
						-- We're holding the button; show the settings.
						game.pauseTimer()
						settings.show()
						game.resumeTimer()
						level.draw()
					end
				elseif control.isTouch() == 1 then
					if touch.click() == 1 and not laser.active then
						x, y = touch.pos()
						if x >= 360 and x < 400 then
							-- We've tapped on the control panel.
							if y >= 78 and y < 156 then
								img = image.load("images/load-pressed.png")
								img:draw(366, 78)
								img:close()
								screen.update()
								sounds.play("clck")
								game.pauseTimer()
								if dialog.show("load-confirmation", { "ok", "cancel" }, "question") == "ok" then
									break
								else
									-- Compensate for the time the dialog was held open.
									game.resumeTimer()
									-- Redraw the level to hide the dialog box.
									level.draw()
									-- Redraw the Load button.
									img = image.load("images/load.png")
									img:draw(366, 78)
									img:close()
								end
							elseif y >= 159 and y < 237 then
								img = image.load("images/restart-pressed.png")
								img:draw(366, 159)
								img:close()
								screen.update()
								sounds.play("clck")
								game.pauseTimer()
								if dialog.show("restart-confirmation", { "ok", "cancel" }, "question") == "ok" then
									game.start()
								else
									-- Compensate for the time the dialog was held open.
									game.resumeTimer()
									-- Redraw the level to hide the dialog box.
									level.draw()
									-- Redraw the Restart button.
									img = image.load("images/restart.png")
									img:draw(366, 159)
									img:close()
								end
							end
						else
							-- We've tapped on the game grid.
							x = math.floor(x / tiles.size)
							y = math.floor(y / tiles.size)
							if x >= 0 and x < level.width and y >= 0 and y < level.height then
								level.click(x, y)
							end
						end
					elseif touch.hold() == 1 and not laser.active then
						x, y = touch.pos()
						if x < 360 then
							-- We've tapped on the game grid.
							x = math.floor(x / tiles.size)
							y = math.floor(y / tiles.size)
							if x >= 0 and x < level.width and y >= 0 and y < level.height then
								local helpPage = tiles.definitions[level.getTile(x, y)].helpPage
								if helpPage then
									-- Show the help page.
									sounds.play("clck")
									game.pauseTimer()
									help.showInGame(helpPage)
									-- Redraw to hid the help page.
									game.drawControls()
									game.drawClock()
									level.draw()
									-- Resume the game.
									game.resumeTimer()
									screen.update()
								else
									-- Help not found, so buzz at the user.
									sounds.play("bzzt")
								end
							end
						end
					end
				end
			end
			
			-- Process game actions.
			laser.update()
			laser.draw()
			screen.update()
			laser.erase()
			level.clearTemporaryTiles()
			laser.visible = true
			
			local newTime = os.time()
			if newTime ~= game.lastUpdateTime then game.lastUpdateTime = newTime game.drawClock() end
			
			if laser.active or not hasInput then
				os.sleep(10)
			end
			
			-- Is the game over?
			if level.irises == 0 and not laser.active then
				-- We completed the level!
				game.drawClock()
				level.setTimeTaken(game.levelFile, game.levelNumber, game.lastDrawnClockSeconds)
				dialog.show("level-completed", { "ok" }, "exclamation")
				game.levelNumber = game.levelNumber + 1
				if game.levelNumber > level.getLevelCount(game.levelFile) then
					break
				else
					game.start()
				end
			end
		end
	end,
	drawLevelSelectionNumber = function(number, x, y)
		local digits = 1
		if number > 9 then digits = digits + 1 end
		if number > 99 then digits = digits + 1 end
		x = x + digits * 3
		for digitIndex = 1, digits do
			game.levelSelectDigits:draw(x - (number % 10) * 6, y, x, y, 6, 10)
			number = math.floor(number / 10)
			x = x - 6
		end
	end,
	closeLevelSelection = function()
		local buttonPressed = image.load("images/return-to-main-menu-pressed.png")
		buttonPressed:draw(366, 36)
		buttonPressed:close()
		screen.update()
		-- Play a sound.
		sounds.play("clck")
		-- Signify that we've quit and set a dummy level number.
		game.quit = true
		game.levelNumber = -1
	end,
	drawLevelSelectionButtons = function()
		
		-- Clear the space behind the buttons.
		screen.fillrect(0, 0, 360, 240, color.new(0, 0, 0))
		
		-- How many levels are there?
		local levelCount = level.getLevelCount(game.levelFile)
		
		-- Draw the buttons.
		local buttonTicked = image.load("images/level-select-cross.png")
		local buttonCrossed = image.load("images/level-select-tick.png")
		for y = 0, 3 do
			for x = 0, 3 do
				levelNumber = game.levelOffset + x + y * 4 + 1
				if levelNumber > 0 and levelNumber <= levelCount then
					if game.levelTimes[levelNumber] then
						buttonCrossed:draw(x * 90 + 1, y * 60 + 2)
					else
						buttonTicked:draw(x * 90 + 1, y * 60 + 2)
					end
					game.drawLevelSelectionNumber(levelNumber, x * 90 + 60, y * 60 + 25)
				end
			end
		end
		buttonTicked:close()
		buttonCrossed:close()

		-- Draw the navigation buttons.
		if game.levelOffset > 0 then
			img = image.load("images/scroll-up.png")
		else
			img = image.load("images/scroll-up-disabled.png")
		end
		img:draw(366, 3)
		img:close()
		
		if (game.levelOffset + 16) < levelCount then
			img = image.load("images/scroll-down.png")
		else
			img = image.load("images/scroll-down-disabled.png")
		end
		img:draw(366, 207)
		img:close()
		
	end,
	showLevelSelection = function()
	
		-- Load the level times.
		game.levelTimes = level.getTimesTaken(game.levelFile)
		
		-- Draw the sidebar.
		screen.fillrect(360, 0, 40, 240, color.new(192, 192, 192))
		screen.fillrect(360, 0, 2, 240, color.new(255, 255, 255))
		
		-- Draw the main menu button.
		img = image.load("images/return-to-main-menu.png")
		img:draw(366, 36)
		img:close()
		
		game.levelNumber = nil
		game.levelSelectDigits = image.load("images/level-select-digits.png")
		
		while not game.levelNumber do
			
			game.drawLevelSelectionButtons()
			
			screen.update()
			
			local levelCount = level.getLevelCount(game.levelFile)
			
			-- Handle input.
			while true do
			
				-- Attempt to get input.
				if control.read() == 1 then
					if control.isButton() == 1 then
						if button.click() == 1 then
							game.closeLevelSelection()
							break
						elseif button.hold() == 1 then
							-- We're holding the button; show the settings.
							settings.show()
							break
						end
					elseif control.isTouch() == 1 then
						if touch.click() == 1 then
							x, y = touch.pos()
							if x >= 360 and x < 400 then
								-- We've tapped on the control panel.
								if y >= 36 and y < 204 then
									-- We've tapped "Return to main menu".
									game.closeLevelSelection()
									break
								elseif y >= 3 and y < 33 and game.levelOffset > 0 then
									-- Scroll to the next set of levels.
									game.levelOffset = game.levelOffset - 16
									-- Draw the pressed button and play a sound.
									img = image.load("images/scroll-up-pressed.png")
									img:draw(366, 3)
									img:close()
									screen.update()
									sounds.play("clck")
									os.sleep(10)
									-- Draw the new selection screen.
									game.drawLevelSelectionButtons()
									screen.update()
								elseif y >= 207 and y < 237 and (game.levelOffset + 16) < levelCount then
									-- Scroll to the next set of levels.
									game.levelOffset = game.levelOffset + 16
									-- Draw the pressed button and play a sound.
									img = image.load("images/scroll-down-pressed.png")
									img:draw(366, 207)
									img:close()
									screen.update()
									sounds.play("clck")
									os.sleep(10)
									-- Draw the new selection screen.
									game.drawLevelSelectionButtons()
									screen.update()
								end
							else
								-- We've tapped on a level selection button.
								x = math.floor((x - 1) / 90)
								y = math.floor((y - 2) / 60)
								if x >= 0 and x < 4 and y >= 0 and y < 4 then
									local levelNumber = game.levelOffset + x + y * 4 + 1
									if levelNumber > 0 and levelNumber <= levelCount then
										-- Set the selected level number.
										game.levelNumber = levelNumber
										-- Draw the pressed button graphic.
										local buttonPressed
										if game.levelTimes[levelNumber] then
											buttonPressed = image.load("images/level-select-tick-pressed.png")
										else
											buttonPressed = image.load("images/level-select-cross-pressed.png")
										end
										buttonPressed:draw(x * 90 + 1, y * 60 + 2)
										buttonPressed:close()
										game.drawLevelSelectionNumber(levelNumber, x * 90 + 61, y * 60 + 26)
										screen.update()
										-- Play a sound.
										sounds.play("clck")
										break
									end
								end
							end
						end
					end
				else
					os.wait(10)
				end
				
			end
		end
		
		game.levelSelectDigits:close()
		game.levelSelectDigits = nil
		
		if game.levelNumber < 0 then game.levelNumber = nil end
		
		os.wait(10)
		
		game.levelTimes = nil
	end
}